
import pygments
from typing import Optional
from whats_that_code.election import guess_language_all_methods
from pygments.lexers import TextLexer, get_lexer_by_name
from pygments.lexers import guess_lexer as pygments_guess_lexer
from pygments.lexer import Lexer
from pygments.formatter import Formatter
from pygments.formatters import  Terminal256Formatter
from pygments.util import ClassNotFound

def guess_lexer(code: str) -> Optional[Lexer]:
    language_name = guess_language_all_methods(code)
    if language_name:
        try:
            return get_lexer_by_name(language_name)
        except ClassNotFound:
            pass
    try:
        return pygments_guess_lexer(code)
    except Exception:
        pass

def highlight_code(code: str, lexer: Optional[Lexer] = None, formatter: Optional[Formatter] = None):
    if lexer is None:
        lexer = guess_lexer(code) or TextLexer()
    if formatter is None:
        formatter = Terminal256Formatter()
    highlighted_code = pygments.highlight(code, lexer, formatter)
    return highlighted_code



import click, sys
from .utils import get_version
from .commands import assistant, scan, thread
from .settings import Settings

VERSION = get_version()
CONTEXT_SETTINGS = {"max_content_width": sys.maxsize}

@click.group(context_settings = CONTEXT_SETTINGS)
@click.version_option(str(VERSION), "-v", "--version")
@click.option("--debug", is_flag = True, hidden = True, help = "Enable debug mode")
@click.pass_context
def cli(ctx: click.Context, debug: bool):
    """A tool to use AI to analyze and interact with vectorized data from custom sources of information."""
    ctx.ensure_object(dict)
    ctx.obj["DEBUG"] = debug

def register_commands():
    cli.add_command(scan)
    cli.add_command(assistant)
    cli.add_command(thread)

def main():
    Settings.init()
    register_commands()
    cli()



import validators, click, requests, time
from openai.types.beta import VectorStore
from typing import Tuple, Dict, Optional
from prompt_toolkit import prompt
from pathlib import Path
from datetime import datetime, timezone
from click import types as ctypes
from .. import ai, utils
from ..api_objects import Assistant, Thread, CONVERSATION_INITS, ConversationInit
from ..settings import Settings
from ..web import process_url
from ..files.processing import process_file, process_dir
from ..files import cache as FileCache
from ..errors import NotSupportedError
from ..data import DataUnit

@click.command()
@click.argument("user_input", nargs = -1)
@click.option("-tn", "--thread_name", help = "The name of the thread. [Optional: can't be restored if not specified.]", default = "")
@click.option("-sm", "--send_messages", type = ctypes.BOOL, default = "false", help = "Send content directly in messages alongside the VectorStore & FileSearch tool.\nThis increases the API cost, and is recommended to be used alongside saving/restoring threads with the '--thread_name' option for larger amounts of content such as codebases.")
@click.pass_context
def scan(ctx: click.Context, user_input: Tuple[str, ...], thread_name: str, send_messages: bool):
    """Scan and process the given input (URL or file path), and offer an interactive prompt to inquire about the vectorized data."""
    settings = Settings() # type: ignore
    FileCache.init()
    debug = ctx.obj.get("DEBUG", False)

    while True:
        # prompt user for data source if not provided as an argument
        if user_input:
            input_str = "".join(user_input)
        else:
            input_str = prompt("Enter a URL or local file path: ").strip('\'"')

        # invoke process_input func to handle processing of data and retrieve vector store/file id(s)
        try:
            resources = process_input(input_str)
            break
        except NotSupportedError as ex:
            print(ex)
            user_input = ()
            continue
        except (requests.RequestException, requests.HTTPError) as ex:
            print(f"An error occurred while sending a web request to the provided URL: {utils.ex_to_str(ex, include_traceback = debug)}")
            user_input = ()
            continue
        except Exception as ex:
            print(f"An unhandled error occurred while processing input: {utils.ex_to_str(ex, include_traceback = debug)}")
            user_input = ()
            continue

    # make sure the VectorStore ID we got seems correct
    vector_store_id = resources.vector_store_id
    try:
        assert len(vector_store_id) > 0, "empty"
        assert vector_store_id.startswith("vs_"), "invalid format"
    except AssertionError as ex:
        print(f"An unknown occurred while processing input: invalid VectorStore ID: {utils.ex_to_str(ex, include_traceback = debug)}")
        return

    # make sure thread name isn't already taken
    if thread_name:
        # allow user to fix thread name if it's already taken
        thread = settings.threads.get_by_name(thread_name)
        if thread:
            print("The thread name you provided has already been used. Please select a different name.")
            taken_names = [t.name for t in settings.threads]
            thread_name = prompt("Thread name: ", validator = utils.ChoiceValidator(taken_names, is_blacklist = True, invalid_message = "Thread name is already taken.", case_sensitive = False))
            utils.delete_lines(2)

        if not thread_name:
            print("No new thread name provided. Thread will not be saved.")

    if thread_name:
        print(f"Thread will be saved: {thread_name}")

    assistant: Assistant = settings.assistants[0]
    if len(settings.assistants) > 1:
        # list assistant choices, map numeric index to name
        assistant_choices: Dict[int, Assistant] = {}
        for idx, assistant in enumerate(settings.assistants, start = 1):
            assistant_choices[idx] = assistant
            print(f"{idx}) {assistant.name}")

        # prompt user to select assistant 
        choices = list(assistant_choices.keys())
        choice = int(prompt("Select an assistant: ", validator = utils.NumericChoiceValidator(choices)))
        assistant = assistant_choices[choice]

        # remove assistant selection menu, output valid choice
        utils.delete_lines(len(assistant_choices) + 1)
        print(f"Using selected assistant: {assistant.name}")

    # verify vector store validity w/ openai, output some generic info
    processing: bool = True
    vector_store: Optional[VectorStore] = None
    while processing:
        try:
            first_run = vector_store is None
            vector_store = ai.Client.beta.vector_stores.retrieve(vector_store_id) # model dump example: https://pastebin.com/k4fwANdi
            assert vector_store.status != "expired", "VectorStore is expired."

            id, name = vector_store_id, vector_store.name
            total_files = vector_store.file_counts.total
            completed_files = vector_store.file_counts.completed
            pending_files = vector_store.file_counts.in_progress
            processing = vector_store.status != "completed"

            size_str = DataUnit.bytes_to_str(vector_store.usage_bytes)
            info = f"{id} ({name}), {completed_files}/{total_files} file(s), {size_str}"
            if first_run: 
                print(f"Successfully established VectorStore!", end = " ")
                if processing:
                    print(f"Processing data from {pending_files} file(s)...")
                else:
                    print(f"{completed_files}/{total_files} file(s) have already been processed.")

            if processing: 
                time.sleep(2)
                continue

            print(f"VectorStore ready for use. [{info}]")
        except Exception as ex:
            print(f"Failed to validate VectorStore from provided ID ({vector_store_id}): {utils.ex_to_str(ex, include_traceback = debug)}")
            return

    # create thread for this conversation
    try:
        api_thread = ai.create_thread(
            resources,
            file_search = assistant.file_search,
            code_interpreter = assistant.interpret_code,
            send_messages = send_messages
        )
        print(f"Thread created with ID: {api_thread.id}")
    except Exception as ex:
        print(f"Error creating thread: {utils.ex_to_str(ex, include_traceback = debug)}")
        return

    # create internal representation of the thread (simplified)
    thread = Thread(
        id = api_thread.id, 
        name = thread_name,
        assistant = (assistant.name, assistant.id),
        created_at = datetime.fromtimestamp(api_thread.created_at, timezone.utc) 
    )

    # saved thread if a name was provided to identify it (so it can be restored later)
    if len(thread.name):
        settings.threads.append(thread)
        settings.save()

    # initialize the conversation with a summary
    try:
        default_ci = ConversationInit(
            user_msg = "Generating summary of content...",
            gpt_msg = "Please identify what the provided content is and provide a summary."
        )
        ci = CONVERSATION_INITS.get(assistant, default_ci)
        print(ci.user_msg)
        ai.get_thread_response(thread.id, assistant.id, ci.gpt_msg, auto_print = True)
    except Exception as ex:
        print(f"Error initializing conversation: {utils.ex_to_str(ex, include_traceback = debug)}")
        return
    
    print("\nYou can now ask questions about the content. Type 'exit' to quit.")
    while True:
        input_str = prompt("\nyou > ")
        utils.conditional_exit(input_str)
        try:
            ai.get_thread_response(thread.id, assistant.id, input_str, auto_print = True)
        except Exception as ex:
            print(f"\nError occurred during conversation: {utils.ex_to_str(ex, include_traceback = debug)}")

def process_input(user_input: str) -> ai.Resources:
    """Takes user input, attempts to return OpenAI VectorStore ID after processing data."""
    utils.conditional_exit(user_input)

    path = Path(user_input)
    if path.exists():
        if path.is_file():
            return process_file(path)
        elif path.is_dir():
            return process_dir(path)

    if validators.url(user_input):
        return process_url(user_input)

    raise NotSupportedError()



import inspect, hashlib, xxhash
from enum import Enum
from typing import Any, List, Union, NamedTuple, cast
from types import ModuleType
from pathlib import Path
from .errors import ValueTypeError

class DataMode(Enum):
    JSON = "json"
    BIN = "binary"

    def ext(self) -> str:
        extensions = {
            DataMode.JSON: "json",
            DataMode.BIN: "bin"
        }
        try:
            return extensions[self]
        except KeyError:
            raise ValueError(f"Unsupported DataMode: {self}")

class DataUnit:
    """
    Utility class for converting sizes in bytes to human-readable string representations with appropriate units.
    
    Attributes:
        B (int): Size multiplier for bytes.
        KB (int): Size multiplier for kilobytes.
        MB (int): Size multiplier for megabytes.
        GB (int): Size multiplier for gigabytes.
        TB (int): Size multiplier for terabytes.

    Methods:
        _get_units(): Retrieve a list of available unit names based on class attributes.
        bytes_to_str(sz_bytes: int) -> str: Convert a size in bytes to a human-readable string representation.
    """
    __excludes__ = ["units"]

    B: int = 1
    KB: int = 2 ** 10
    MB: int = 2 ** 20
    GB: int = 2 ** 30
    TB: int = 2 ** 40

    units: List[str] = []

    @staticmethod
    def _get_units() -> List[str]:
        # "B", "KB", "MB", "GB", "TB"
        if len(DataUnit.units):
            return DataUnit.units
        units = [
            (name, value) for name, value in inspect.getmembers(DataUnit) 
            if not name.startswith("__") 
            and name not in DataUnit.__excludes__ 
            and not callable(value)
        ]
        units = sorted(units, key=lambda x: x[1])
        DataUnit.units = [name for name, _ in units]
        return DataUnit.units

    @staticmethod
    def bytes_to_str(sz_bytes: int) -> str:
        """
        Convert a size in bytes to a human-readable string representation.

        Parameters:
            sz_bytes (int): The size in bytes to convert to a human-readable string.

        Returns:
            str: A human-readable string representation of the size, including units (e.g. "1.23 MB").
        """
        assert sz_bytes >= 0, "sz_bytes must be non-negative"
        size, uidx = sz_bytes, 0
        units = DataUnit.units or DataUnit._get_units()
        while size >= DataUnit.KB and uidx < len(units) - 1:
            size /= float(DataUnit.KB)
            uidx += 1
        return f"{size:.2f} {units[uidx]}"

class _HashAlg(NamedTuple):
    module: ModuleType
    function_name: str

    def init(self) -> Any:
        alg_init = getattr(self.module, self.function_name)
        return alg_init()

class HashAlg(Enum):
    """An enumeration class representing various hashing algorithms and offering their implementation."""
    SHA_256 = _HashAlg(hashlib, "sha256")
    SHA3_256 = _HashAlg(hashlib, "sha3_256")
    XXH_32 = _HashAlg(xxhash, "xxh32")
    XXH_64 = _HashAlg(xxhash, "xxh64")
    XXH_128 = _HashAlg(xxhash, "xxh128")
    XXH3_128 = _HashAlg(xxhash, "xxh3_128")
    XXH3_64 = _HashAlg(xxhash, "xxh3_64")

    def calculate(
        self,
        _input: Union[Path, str, bytes], 
        intdigest: bool = False
    ) -> Union[str, int]:
        """
        Calculate the hash of the input using the specified algorithm.

        Parameters:
            _input (Union[Path, str, bytes]): The input data to calculate the hash for.
            intdigest (bool): Whether to return the hash as an integer or a string.

        Returns:
            Union[str, int]: The calculated hash, either as a string or an integer.

        Raises:
            ValueError: If the hash algorithm/object is not valid.
            ValueTypeError: If the input data type is not one of Path, str, or bytes.
        """
        try:
            alg = cast(_HashAlg, self.value)
            hash_obj = alg.init()
            assert HashAlg.hash_obj_valid(hash_obj)
        except (AssertionError, AttributeError) as ex:
            raise ValueError("Invalid hash object.") from ex

        if isinstance(_input, (bytes, str)):
            _input = _input.encode() if isinstance(_input, str) else _input
            hash_obj.update(_input)
        elif isinstance(_input, Path):
            with open(_input, "rb") as f:
                for chunk in iter(lambda: f.read(8 * DataUnit.KB), b""):
                    hash_obj.update(chunk)
        else:
            raise ValueTypeError(_input, (Path, str, bytes))

        return (
            hash_obj.hexdigest() if not intdigest else
            int.from_bytes(hash_obj.digest(), byteorder = "big")
        )

    @staticmethod
    def hash_obj_valid(alg: Any) -> bool:
        """Check if the given algorithm object is a valid hash algorithm."""
        updater = getattr(alg, "update", None)
        digester = getattr(alg, "digest", None)
        hexdigester = getattr(alg, "hexdigest", None)
        return (
            updater is not None and callable(updater) and
            digester is not None and callable(digester) and
            hexdigester is not None and callable(hexdigester)
        )



import re, pickle, gzip, chardet
from typing import TypeVar, Type, List, Optional, NamedTuple, Union
from pathlib import Path
from .encodings import Encoding
from ..data import DataUnit

T = TypeVar("T")

def write_str(file_path: Path, text: str, compress: bool = False) -> None:
    file_path.parent.mkdir(parents = True, exist_ok = True)
    data = text.encode("utf-8")
    write_bytes(file_path, data, compress)

def read_str(file_path: Path) -> str:
    data = read_bytes(file_path)
    text = data.decode("utf-8")
    return text

def write_bytes(file_path: Path, data: bytes, compress: bool = False) -> None:
    file_path.parent.mkdir(parents = True, exist_ok = True)
    if compress:
        if file_path.suffix != ".bin" and ".gz" not in file_path.suffixes:
            file_path = file_path.with_suffix(file_path.suffix + ".gz")
        data = gzip.compress(data)
    with open(file_path, 'wb') as file:
        file.write(data)

def read_bytes(file_path: Path) -> bytes:
    with open(file_path, 'rb') as file:
        data = file.read()
    if ".gz" in file_path.suffixes or ".bin" in file_path.suffixes:
        try:
            data = gzip.decompress(data)
        except gzip.BadGzipFile:
            pass
    return data

def save_object(file_path: Path, obj: object, compress: bool = True) -> None:
    data = pickle.dumps(obj)
    write_bytes(file_path, data, compress)

def load_object(file_path: Path, cls: Type[T]) -> T:
    obj = load_object_any(file_path)
    if not isinstance(obj, cls):
        raise TypeError(f"Expected object of type {cls.__name__}, but got {type(obj).__name__}")
    return obj

def load_object_any(file_path: Path) -> object:
    data = read_bytes(file_path)
    return pickle.loads(data)

def list_files(directory: Path, recursive: bool = True) -> List[Path]:
    assert directory.is_dir(), f"The provided path to 'list_files' must be a directory. (Value: '{directory}')"
    files = []
    for item in directory.iterdir():
        if item.is_file():
            files.append(item)
        elif item.is_dir() and recursive:
            files.extend(list_files(item, recursive = True))
    return files

def get_encoding(file_path: Path) -> Optional[Encoding]:
    """
    Attempts to determine the encoding of a file based on the first 4 KB. If none is recognized, the function returns 'None'.
    Supported encodings: https://link.justin.ooo/chardet-encodings
    """
    with file_path.open('rb') as f:
        raw_data = f.read(4 * DataUnit.KB) # read the first 4KB of the file
        result = chardet.detect(raw_data)
        encoding_str = result.get("encoding")
        if isinstance(encoding_str, str):
            encoding = Encoding.from_string(encoding_str)
            return encoding
    return None

def has_parent_directory(path: Path, dir_name: str) -> bool:
    normalize = lambda s: f"/{s.strip('/\\').replace('\\', '/')}/"
    dir_name = normalize(dir_name)
    path_str = normalize(str(path))
    return dir_name in path_str

def matches_pattern(text: str, pattern: str) -> bool:
    return re.search(pattern, text) is not None

class FilteredFiles(NamedTuple):
    files: List[Path]
    total_count: int
    valid_count: int

def filter_files(all_files: List[Path]) -> FilteredFiles:
    encoding_whitelist = [Encoding.UTF_8, Encoding.ASCII]
    dir_blacklist = [".git", "node_modules", "site-packages", ".mypy_cache"]
    pattern_blacklist = [r".*\.egg-info"]
    extension_whitelist = {
        '.py', '.js', '.txt', '.md', '.html', '.css', '.java', '.c', '.cpp',
        '.rb', '.php', '.ts', '.json', '.xml', '.csv', '.xlsx', '.pptx', '.docx',
        '.jpg', '.jpeg', '.png', '.gif', '.webp', '.pdf', '.zip', '.tar', '.tex'
    }

    files = [
        file_path for file_path in all_files
        if file_path.suffix in extension_whitelist \
        and get_encoding(file_path) in encoding_whitelist \
        and not any(has_parent_directory(file_path, dir) for dir in dir_blacklist)
    ]
    
    # NOTE(justin): run actual path pattern checks in a separate loop (for optimization purposes)
    # the goal is to filter out *most* invalid paths without regex prior to this using has_parent_directory
    files = [
        file for file in files if \
        not any(matches_pattern(str(file), pattern) for pattern in pattern_blacklist)
    ]

    return FilteredFiles(
        files = files,
        total_count = len(all_files),
        valid_count = len(files)
    )



from dataclasses import dataclass, asdict
from typing import  List, Set, Tuple, Iterable, Optional, Callable, TypeVar, ClassVar, Generic, Protocol
from datetime import datetime
from .. import utils
from ..errors import MultipleObjectsFoundError, MissingSortKeyError

@dataclass
class BaseApiObj:
    def to_dict(self) -> dict:
        obj = asdict(self)
        return utils.convert_datetimes(obj, converter = utils.converter_ts_int)

class ApiObjItem(Protocol):
    id: str
    name: str
    created_at: datetime

    def to_dict(self) -> dict:
        ...

AOT = TypeVar("AOT", bound = ApiObjItem)
T = TypeVar("T")

class ApiObjList(List[AOT], Generic[AOT]):

    API_OBJ_PREFIXES: ClassVar[Set[str]] = {"asst", "thread"}

    def __init__(
        self, 
        items: Iterable[AOT] = [], 
        sort: bool = False,
        key: Optional[Callable[[AOT], T]] = None,
        reverse: bool = False,
        cls: Optional[Callable[..., AOT]] = None,
    ):  
        self._cls = cls

        if not sort:
            # sort by "created_at" by default
            items = sorted(items, key=lambda item: item.created_at)  # type: ignore
            super().__init__(items) # type: ignore
            return

        if not key or not callable(key):
            raise MissingSortKeyError(f"Can't initialize '{self.class_name}' in sorted order without callable key.")

        sorted_items = sorted(items, key=key, reverse=reverse)  # type: ignore
        super().__init__(sorted_items)

    def to_dict_list(self) -> List[dict]:
        return [item.to_dict() for item in self]  # type: ignore

    @staticmethod
    def from_dict_list(objects: List[dict], cls: Callable[..., AOT], **kwargs) -> "ApiObjList[AOT]":
        items = [cls(**obj) for obj in objects]
        return ApiObjList(items, cls = cls, **kwargs)

    def list_by_name(self, name: str, case_sensitive: bool = False) -> List[T]:
        t = lambda s: s if case_sensitive else s.lower()  # transform func
        return [item for item in self if t(item.name) == t(name)]  # type: ignore

    def get_by_name(self, name: str, default: Optional[T] = None, case_sensitive: bool = False) -> Optional[T]:
        items = self.list_by_name(name, case_sensitive=case_sensitive)
        if len(items) == 0:
            return default
        if len(items) > 1:
            assert self._cls
            raise MultipleObjectsFoundError(self.class_name, "name", name)
        return items[0]

    def get_by_id(self, id: str) -> Tuple[Optional[AOT], int]:
        for idx, item in enumerate(self):
            if item.id == id:  # type: ignore
                return item, idx
        return None, -1
    
    def get(self, identifier: str) -> Tuple[Optional[AOT], int]:
        """
        Get an API object from an identifier.

        This method retrieves an item object based on the provided identifier. The identifier can be one of the following:
        - API ID: If the identifier string seems to be an API ID, it is treated as such and the item is retrieved using this ID.
        - Item name: If the identifier is a valid item name, the item is retrieved based on the name.
        - Menu ID: If the identifier is numeric, it retrieves the item associated with that number in the 'list' command. (index + 1)

        Parameters:
            identifier (str): The identifier used to retrieve the item.

        Returns:
            Tuple[Optional[AOT], int]: A tuple containing the retrieved API object representation and its index in the list of items. Returns (None, -1) if not found.
        """
        idx = -1

        if any(identifier.startswith(f"{p}_") for p in ApiObjList.API_OBJ_PREFIXES):
            _, idx = self.get_by_id(identifier)
            if idx == -1:
                return None, -1

        if idx == -1:
            item = self.get_by_name(identifier)
            if item:
                _, idx = self.get_by_id(item.id)  # type: ignore

        if idx == -1:
            idx = utils.try_parse_int(identifier)
            idx = -1 if idx is None else idx - 1

        if idx < 0 or idx > len(self) - 1:
            return None, -1

        item = self[idx]
        return item, idx
    
    @property
    def class_name(self):
        return self._cls.__name__ if self._cls else "Unknown"



from dataclasses import dataclass, field
from datetime import datetime
from typing import Iterable, Tuple
from openai.types.beta import Thread as APIThread
from .. import utils, ai
from .generic import ApiObjList, BaseApiObj

@dataclass
class Thread(BaseApiObj):
    id: str
    name: str
    assistant: Tuple[str, str] # name, id
    created_at: datetime = field(default_factory = utils.utc_now)

    def get_api_obj(self) -> APIThread:
        if not self.id:
            raise ValueError("Thread ID must be specified to retrieve API object.")
        api_thread = ai.get_thread(self.id)
        return api_thread

class ThreadList(ApiObjList[Thread]):

    def __init__(self, threads: Iterable[Thread] = [], **kwargs):
        super().__init__(threads, cls = Thread, **kwargs)

    @staticmethod
    def from_dict_list(threads: Iterable[Thread], **kwargs) -> "ThreadList": # type: ignore
        return ApiObjList.from_dict_list(threads, cls = Thread, **kwargs) # type: ignore
